import pytest

from secedgar.core.rest import (_combine_dicts, get_company_concepts,
                                get_company_facts, get_submissions,
                                get_xbrl_frames)


class TestRest:
    @pytest.mark.parametrize("dicts,expected",
                             [
                                 ([{"A": [1, 2, 3], "B": [4, 5, 6]},
                                  {"A": [7, 8], "B": [0, 1, 2]}],
                                  {'A': [1, 2, 3, 7, 8], 'B': [4, 5, 6, 0, 1, 2]}),
                                 ([{"A": [1, 2, 3]},
                                  {"B": [4, 5, 6]}],
                                  {"A": [1, 2, 3], "B": [4, 5, 6]}),
                                 ([{"A": [1]}],
                                  {"A": [1]})
                             ]
                             )
    def test__combine_dicts(self, dicts, expected):
        assert _combine_dicts(*dicts) == expected

    @pytest.mark.parametrize(
        "recent",
        [True, False]
    )
    @pytest.mark.smoke
    def test_get_submissions(self, mock_user_agent, recent):
        submissions = get_submissions(lookups=["aapl"],
                                      user_agent=mock_user_agent,
                                      recent=recent)
        assert submissions
        # Make sure Apple's CIK shows up properly
        assert str(submissions["aapl"]["cik"]) == "320193"

        # Make sure there are accession numbers
        assert submissions["aapl"]["filings"]["recent"]["accessionNumber"]

        # Result should be dictionary
        assert isinstance(submissions, dict)

    @pytest.mark.smoke
    def test_get_company_concepts(self, mock_user_agent):
        concept = "AccountsPayableCurrent"
        concepts = get_company_concepts(lookups=["AAPL"],
                                        user_agent=mock_user_agent,
                                        concept_name=concept)
        assert concepts
        # Ensure CIK is correct
        assert str(concepts["AAPL"]["cik"]) == "320193"

        # Make sure that there are results for accounts payable
        assert concepts["AAPL"]["units"]["USD"]

        # Result should be dictionary
        assert isinstance(concepts, dict)

    @pytest.mark.smoke
    def test_get_company_facts(self, mock_user_agent):
        facts = get_company_facts(lookups=["aapl"], user_agent=mock_user_agent)

        assert facts
        # Ensure CIK is correct - sometimes will give number, so cast to string
        assert str(facts["aapl"]["cik"]) == "320193"

        # Make sure there are facts
        assert facts["aapl"]["facts"]

        # Make sure that us-gaap and dei both are keys
        assert "us-gaap" in facts["aapl"]["facts"]
        assert "dei" in facts["aapl"]["facts"]

        # Make sure we can get Revenues for Apple
        assert facts["aapl"]["facts"]["us-gaap"]["Revenues"]["units"]["USD"]

        # Result should be dictionary
        assert isinstance(facts, dict)

    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "concept,instantaneous",
        [
            ("Revenues", False),
            ("Revenues", True)
        ]
    )
    def test_get_xbrl_frames(self, mock_user_agent, concept, instantaneous):
        frames = get_xbrl_frames(user_agent=mock_user_agent,
                                 concept_name=concept,
                                 year=2020,
                                 quarter=1,
                                 instantaneous=instantaneous)
        # Check to make sure we got the right frame
        assert frames["tag"] == concept

        # Make sure there is data
        assert frames["data"]

        # Data should have accn and cik as keys
        assert "accn" in frames["data"][0]
        assert "cik" in frames["data"][0]

        assert isinstance(frames, dict)



class EDGARQueryError(Exception):
    """This error is thrown when a query receives a response that is not a 200 response."""
    pass


class CIKError(Exception):
    """This error is thrown when an invalid CIK is given."""
    pass


class FilingTypeError(Exception):
    """This error is thrown when an invalid filing type is given."""
    pass


class NoFilingsError(Exception):
    """This error is thrown when no filings are found."""
    pass



__version__ = '0.6.0'

from secedgar.core.combo import ComboFilings  # noqa: F401
from secedgar.core.company import CompanyFilings  # noqa: F401
from secedgar.core.daily import DailyFilings  # noqa: F401
from secedgar.core.filing_types import FilingType  # noqa: F401
from secedgar.core.filings import filings  # noqa: F401
from secedgar.core.quarterly import QuarterlyFilings  # noqa: F401



import os
from datetime import date, datetime

import pytest
import secedgar.utils as utils
from secedgar.client import NetworkClient
from secedgar.core.daily import DailyFilings
from secedgar.tests.utils import MockResponse, datapath

cik_file_pairs = [("1000228", "0001209191-18-064398.txt"),
                  ("1000275", "0001140361-18-046093.txt"),
                  ("1000275", "0001140361-18-046095.txt"),
                  ("1000694", "0001144204-18-066755.txt"),
                  ("1001085", "0001104659-18-075315.txt")]


@pytest.fixture(scope="module")
def mock_daily_quarter_directory(monkeymodule):
    """Mocks directory of all daily filings for quarter."""
    monkeymodule.setattr(
        DailyFilings, "_get_listings_directory",
        MockResponse(
            datapath_args=["filings", "daily", "daily_index_2018_QTR4.htm"]))


@pytest.fixture(scope="module")
def mock_daily_idx_file(monkeymodule):
    """Mock idx file from DailyFilings."""

    def _mock_daily_idx_file(*args, **kwargs):
        with open(datapath("filings", "daily", "master.20181231.idx")) as f:
            return f.read()

    monkeymodule.setattr(DailyFilings, "_get_master_idx_file",
                         _mock_daily_idx_file)


class TestDaily:

    @pytest.mark.parametrize("date,expected", [(date(2020, 1, 1), 1),
                                               (date(2020, 3, 31), 1),
                                               (date(2020, 4, 1), 2),
                                               (date(2020, 6, 30), 2),
                                               (date(2020, 7, 1), 3),
                                               (date(2020, 9, 30), 3),
                                               (date(2020, 10, 1), 4),
                                               (date(2020, 12, 31), 4)])
    def test_quarter(self, date, expected, mock_user_agent):
        assert DailyFilings(date=date, user_agent=mock_user_agent).quarter == expected

    @pytest.mark.parametrize("date,expected_filename", [
        (date(2020, 1, 1), "master.20200101.idx"),
        (date(2020, 3, 31), "master.20200331.idx"),
        (date(2020, 4, 1), "master.20200401.idx"),
        (date(2020, 6, 30), "master.20200630.idx"),
    ])
    def test_idx_filename(self, date, expected_filename, mock_user_agent):
        assert DailyFilings(date=date, user_agent=mock_user_agent).idx_filename == expected_filename

    @pytest.mark.parametrize("bad_date", [1.0, 12, "12/31/2018"])
    def test_bad_date_format_fails(self, bad_date):
        with pytest.raises(TypeError):
            DailyFilings(bad_date)

    @pytest.mark.parametrize("key,url", [(cik,
                                          "http://www.sec.gov/Archives/edgar/data/{cik}/{f}".format(
                                              cik=cik, f=f))
                                         for cik, f in cik_file_pairs])
    def test_get_urls(self, mock_user_agent, mock_daily_quarter_directory, mock_daily_idx_file,
                      key, url):
        daily_filing = DailyFilings(date(2018, 12, 31), user_agent=mock_user_agent)
        assert url in daily_filing.get_urls()[key]

    def test_get_listings_directory(self, mock_user_agent, mock_daily_quarter_directory):
        daily_filings = DailyFilings(date(2018, 12, 31),
                                     user_agent=mock_user_agent)
        daily_filing_listing_directory = daily_filings._get_listings_directory()
        assert daily_filing_listing_directory.status_code == 200
        assert "master.20181231.idx" in daily_filing_listing_directory.text

    @pytest.mark.parametrize("company_name", [
        "HENRY SCHEIN INC", "ROYAL BANK OF CANADA", "NOVAVAX INC",
        "BROOKFIELD ASSET MANAGEMENT INC.",
        "PERUSAHAAN PERSEROAN PERSERO PT TELEKOMUNIKASI INDONESIA TBK"
    ])
    def test_get_master_idx_file(self, mock_user_agent, mock_daily_quarter_directory,
                                 mock_daily_idx_file, company_name):
        daily_filing = DailyFilings(date(2018, 12, 31),
                                    user_agent=mock_user_agent)
        assert company_name in daily_filing._get_master_idx_file()

    @pytest.mark.parametrize("year,month,day,quarter", [(2018, 1, 1, 1),
                                                        (2017, 5, 1, 2),
                                                        (2016, 6, 30, 2),
                                                        (2015, 7, 1, 3),
                                                        (2014, 9, 30, 3),
                                                        (2013, 10, 1, 4),
                                                        (2012, 11, 20, 4),
                                                        (2011, 12, 31, 4)])
    def test_path_property(self, year, month, day, quarter, mock_user_agent):
        daily_filing = DailyFilings(date(year, month, day), user_agent=mock_user_agent)
        assert daily_filing.path == "Archives/edgar/daily-index/{year}/QTR{quarter}/".format(
            year=year, quarter=quarter)

    @pytest.mark.parametrize("date", [
        date(2020, 1, 1),
        date(2020, 3, 30),
        date(1993, 12, 31),
        datetime(2020, 1, 1)
    ])
    def test_good_date_setter(self, date, mock_user_agent):
        daily_filing = DailyFilings(date=date, user_agent=mock_user_agent)
        assert daily_filing.date == date

    @pytest.mark.parametrize("bad_date", [20201231, "20201231", "2020/12/31"])
    def test_bad_date_on_init(self, bad_date):
        with pytest.raises(TypeError):
            _ = DailyFilings(date=bad_date)

    @pytest.mark.parametrize("bad_date", [20201231, "20201231", "2020/12/31"])
    def test_bad_date_setter_after_init(self, bad_date, mock_user_agent):
        daily_filing = DailyFilings(date=date(2020, 1, 1), user_agent=mock_user_agent)
        with pytest.raises(TypeError):
            daily_filing.date = bad_date

    def test_no_params(self, mock_user_agent):
        """Params should always be empty."""
        daily_filing = DailyFilings(date(2020, 1, 1), user_agent=mock_user_agent)
        assert not daily_filing.params

    @pytest.mark.parametrize("date_tuple,formatted", [
        ((1994, 1, 2), "010294"),
        ((1994, 12, 31), "123194"),
        ((1995, 1, 1), "950101"),
        ((1995, 1, 2), "950102"),
        ((1998, 1, 1), "980101"),
        ((1998, 1, 2), "980102"),
        ((1998, 3, 31), "19980331"),
        ((1998, 4, 1), "19980401"),
        ((1999, 1, 1), "19990101"),
    ])
    def test_master_idx_date_format(self, date_tuple, formatted, mock_user_agent):
        daily_filing = DailyFilings(date(*date_tuple), user_agent=mock_user_agent)
        assert daily_filing._get_idx_formatted_date() == formatted

    @pytest.mark.parametrize("cik,file", cik_file_pairs)
    def test_save_default(self, tmp_data_directory,
                          mock_daily_quarter_directory, mock_daily_idx_file,
                          mock_filing_response,
                          mock_user_agent, cik, file):
        daily_filing = DailyFilings(date(2018, 12, 31), user_agent=mock_user_agent)
        daily_filing.save(tmp_data_directory)
        subdir = os.path.join("20181231", cik)
        path_to_check = os.path.join(tmp_data_directory, subdir, file)
        assert os.path.exists(path_to_check)

    @pytest.mark.parametrize("file", [cf[1] for cf in cik_file_pairs])
    def test_save_with_single_level_date_dir_pattern(
            self, tmp_data_directory, mock_user_agent, mock_daily_quarter_directory,
            mock_daily_idx_file, mock_filing_response, file):
        daily_filing = DailyFilings(date(2018, 12, 31), user_agent=mock_user_agent)
        daily_filing.save(tmp_data_directory,
                          dir_pattern="{date}",
                          date_format="%Y-%m-%d")
        path_to_check = os.path.join(tmp_data_directory, "2018-12-31", file)
        assert os.path.exists(path_to_check)

    @pytest.mark.parametrize("cik,file", cik_file_pairs)
    def test_save_with_single_level_cik_dir_pattern(
            self, tmp_data_directory, mock_user_agent, mock_daily_quarter_directory,
            mock_daily_idx_file, mock_filing_response, cik, file):
        daily_filing = DailyFilings(date(2018, 12, 31), user_agent=mock_user_agent)
        daily_filing.save(tmp_data_directory, dir_pattern="{cik}")
        path_to_check = os.path.join(tmp_data_directory, cik, file)
        assert os.path.exists(path_to_check)

    @pytest.mark.parametrize("cik,file", cik_file_pairs)
    def test_save_with_multi_level_dir_pattern(self, tmp_data_directory,
                                               mock_user_agent,
                                               mock_daily_quarter_directory,
                                               mock_daily_idx_file,
                                               mock_filing_response, cik, file):
        daily_filing = DailyFilings(date(2018, 12, 31), user_agent=mock_user_agent)
        daily_filing.save(tmp_data_directory,
                          dir_pattern="{date}/{cik}",
                          date_format="%Y-%m-%d")
        subdir = os.path.join("2018-12-31", cik)
        path_to_check = os.path.join(tmp_data_directory, subdir, file)
        assert os.path.exists(path_to_check)

    @pytest.mark.parametrize("cik,file", cik_file_pairs)
    def test_save_with_multi_level_dir_pattern_date_not_first(
            self, tmp_data_directory, mock_user_agent, mock_daily_quarter_directory,
            mock_daily_idx_file, mock_filing_response, cik, file):
        daily_filing = DailyFilings(date(2018, 12, 31), user_agent=mock_user_agent)
        daily_filing.save(tmp_data_directory,
                          dir_pattern="{cik}/{date}",
                          date_format="%Y-%m-%d")
        subdir = os.path.join(cik, "2018-12-31")
        path_to_check = os.path.join(tmp_data_directory, subdir, file)
        assert os.path.exists(path_to_check)

    @pytest.mark.parametrize("year,quarter", [(1995, 2), (1995, 1), (1994, 4)])
    def test__get_tar_bad_year_quarter(self, year, quarter, mock_user_agent):
        with pytest.raises(ValueError):
            d = DailyFilings(date=date(year, utils.get_month(quarter), 1),
                             user_agent=mock_user_agent)
            d._get_tar_urls()

    def test__get_tar_valid(self, mock_user_agent):
        d = DailyFilings(date=date(2020, 1, 2), user_agent=mock_user_agent)
        urls = d._get_tar_urls()
        assert urls[0].endswith('20200102.nc.tar.gz')
        assert urls[0].startswith('http')

    @pytest.mark.parametrize(
        "bad_entry_filter",
        [
            "entry_filter",
            True,
            False,
            1,
            1.0
        ]
    )
    def test_set_bad_entry_filter(self, bad_entry_filter, mock_user_agent):
        with pytest.raises(ValueError):
            d = DailyFilings(date=date(2020, 1, 2), user_agent=mock_user_agent)
            d.entry_filter = bad_entry_filter

    @pytest.mark.parametrize(
        "good_entry_filter",
        [
            lambda _: True,
            lambda _: False,
            lambda f: f.name.lower().startswith("a")
        ]
    )
    def test_set_good_entry_filter(self, good_entry_filter, mock_user_agent):
        d = DailyFilings(date=date(2020, 1, 2), user_agent=mock_user_agent)
        d.entry_filter = good_entry_filter
        assert callable(d.entry_filter)

    def test_download_all(self, monkeypatch):
        pass

    def test_user_agent_passed_to_client(self, mock_user_agent):
        daily = DailyFilings(date=date(2021, 1, 1), user_agent=mock_user_agent)
        assert daily.client.user_agent == mock_user_agent

    def test_user_agent_client_none(self):
        with pytest.raises(TypeError):
            _ = DailyFilings(date=date(2020, 1, 1),
                             user_agent=None,
                             client=None)

    def test_set_good_client(self, mock_user_agent):
        client = NetworkClient(user_agent=mock_user_agent)
        daily = DailyFilings(date=date(2021, 1, 1), client=client)
        assert daily.client == client



from datetime import date

import pytest
from secedgar.client import NetworkClient
from secedgar.core import DailyFilings, FilingType, QuarterlyFilings
from secedgar.core.combo import ComboFilings
from secedgar.core.filings import filings
from secedgar.exceptions import FilingTypeError


class TestFilings:

    def test_bad_filing_type(self):
        with pytest.raises(FilingTypeError):
            filings(cik_lookup='aapl', filing_type='10-k')

    def test_user_agent_passed_to_objects(self, mock_user_agent):
        f = filings(cik_lookup="aapl",
                    filing_type=FilingType.FILING_10Q,
                    user_agent=mock_user_agent)
        assert f.client.user_agent == mock_user_agent

    def test_client_passed_to_objects(self, mock_user_agent):
        client = NetworkClient(user_agent=mock_user_agent)
        company = filings(cik_lookup="aapl", filing_type=FilingType.FILING_10Q, client=client)
        daily = filings(start_date=date(2021, 1, 1), end_date=date(2021, 1, 1), client=client)
        quarterly = filings(start_date=date(2020, 1, 1), end_date=date(2020, 3, 31), client=client)
        assert company.client == client
        assert daily.client == client
        assert quarterly.client == client

    def test_filing_type_plus_entry_filter_filters_both(self):
        pass

    def test_count_not_implemented(self, mock_user_agent):
        with pytest.raises(NotImplementedError):
            filings(start_date=date(2010, 1, 1),
                    end_date=date(2020, 1, 1),
                    user_agent=mock_user_agent,
                    count=10)

    def test_same_date_returns_daily_filings(self, mock_user_agent):
        day = date(2020, 1, 1)

        f = filings(start_date=day, end_date=day, user_agent=mock_user_agent)
        assert isinstance(f, DailyFilings)
        assert f.date == day
        assert f.year == day.year
        assert f.quarter == day.month // 3 + 1

    def test_exact_quarter_returns_quarterly_filings(self, mock_user_agent):
        f = filings(start_date=date(2020, 1, 1), end_date=date(
            2020, 3, 31), user_agent=mock_user_agent)
        assert isinstance(f, QuarterlyFilings)
        assert f.year == 2020
        assert f.quarter == 1

    def test_mismatch_of_quarters_returns_combo_filings(self, mock_user_agent):
        start_date = date(2020, 1, 1)
        end_date = date(2020, 5, 1)
        f = filings(start_date=start_date, end_date=end_date, user_agent=mock_user_agent)
        assert isinstance(f, ComboFilings)
        assert f.start_date == start_date
        assert f.end_date == end_date

    @pytest.mark.parametrize(
        "kwargs,error",
        [
            (dict(), ValueError),
            (dict(end_date=date(2020, 1, 1)), ValueError),
            (dict(client=NetworkClient(user_agent="Test")), ValueError),
            (dict(entry_filter=lambda f: f.form_type == '10-k'), ValueError),

            # end_date assumed to be today - giving count tries to
            # make daily or quarterly filing, but will not be recognized
            (dict(count=10, client=NetworkClient(user_agent="Test")), NotImplementedError),
            (dict(count=10), NotImplementedError),
        ])
    def test_bad_args_combination_raises_error(self, kwargs, error):
        with pytest.raises(error):
            filings(**kwargs)



import asyncio
import os
import re
import shutil
import tempfile
from abc import abstractmethod
from collections import namedtuple
from queue import Empty, Queue
from threading import Thread

from secedgar.client import NetworkClient
from secedgar.core._base import AbstractFiling
from secedgar.exceptions import EDGARQueryError
from secedgar.utils import make_path


class IndexFilings(AbstractFiling):
    """Abstract Base Class for index filings.

    Args:
        user_agent (Union[str, NoneType]): Value used for HTTP header "User-Agent" for all requests.
            If given None, a valid client with user_agent must be given.
            See the SEC's statement on
            `fair access <https://www.sec.gov/os/accessing-edgar-data>`_
            for more information.
        client (Union[NoneType, secedgar.client.NetworkClient], optional): Client to use for
            fetching data. If None is given, a user_agent must be given to pass to
            :class:`secedgar.client.NetworkClient`.
            Defaults to ``secedgar.client.NetworkClient`` if none is given.
        entry_filter (function, optional): A boolean function to determine
            if the FilingEntry should be kept. E.g. `lambda l: l.form_type == "4"`.
            Defaults to `None`.
        kwargs: Any keyword arguments to pass to ``NetworkClient`` if no client is specified.
    """

    def __init__(self, user_agent=None, client=None, entry_filter=None, **kwargs):
        super().__init__()
        self._client = client or NetworkClient(user_agent=user_agent, **kwargs)
        self._listings_directory = None
        self._master_idx_file = None
        self._filings_dict = None
        self._paths = []
        self._urls = {}
        self.entry_filter = entry_filter

    @property
    def entry_filter(self):
        """A boolean function to be tested on each listing entry.

        This is tested regardless of download method.
        """
        return self._entry_filter

    @entry_filter.setter
    def entry_filter(self, fn):
        if callable(fn):
            self._entry_filter = fn
        else:
            raise ValueError('entry_filter must be a function or lambda.')

    @property
    def client(self):
        """``secedgar.client.NetworkClient``: Client to use to make requests."""
        return self._client

    @property
    def params(self):
        """Params should be empty."""
        return {}

    @property
    @abstractmethod
    def year(self):
        """Passed to children classes."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def quarter(self):
        """Passed to children classes."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def idx_filename(self):
        """Passed to children classes."""
        pass  # pragma: no cover

    @abstractmethod
    def _get_tar_urls(self):
        """Passed to child classes."""
        pass  # pragma: no cover

    @property
    def tar_path(self):
        """str: Tar.gz path added to the client base."""
        return "Archives/edgar/Feed/{year}/QTR{num}/".format(year=self.year,
                                                             num=self.quarter)

    def _get_listings_directory(self, update_cache=False, **kwargs):
        """Get page with list of all idx files for given date or quarter.

        Args:
            update_cache (bool, optional): Whether quarterly directory should update cache. Defaults
                to False.
            kwargs: Any keyword arguments to pass to the client's `get_response` method.

        Returns:
            response (requests.Response): Response object from page with all idx files for
                given quarter and year.
        """
        if self._listings_directory is None or update_cache:
            self._listings_directory = self.client.get_response(
                self.path, self.params, **kwargs)
        return self._listings_directory

    def _get_master_idx_file(self, update_cache=False, **kwargs):
        """Get master file with all filings from given date.

        Args:
            update_cache (bool, optional): Whether master index should be updated
                method call. Defaults to False.
            kwargs: Keyword arguments to pass to
                ``secedgar.client.NetworkClient.get_response``.

        Returns:
            text (str): Idx file text.

        Raises:
            EDGARQueryError: If no file of the form master.<DATE>.idx
                is found.
        """
        if self._master_idx_file is None or update_cache:
            if self.idx_filename in self._get_listings_directory().text:
                master_idx_url = "{path}{filename}".format(
                    path=self.path, filename=self.idx_filename)
                self._master_idx_file = self.client.get_response(
                    master_idx_url, self.params, **kwargs).text
            else:
                raise EDGARQueryError("""File {filename} not found.
                                     There may be no filings for the given day/quarter."""
                                      .format(filename=self.idx_filename))
        return self._master_idx_file

    def get_filings_dict(self, **kwargs):
        """Get all filings inside an idx file.

        Args:
            kwargs: Any kwargs to pass to _get_master_idx_file. See
                ``secedgar.core.daily.DailyFilings._get_master_idx_file``.
        """
        idx_file = self._get_master_idx_file(**kwargs)
        # Will have CIK as keys and list of FilingEntry namedtuples as values
        self._filings_dict = {}
        FilingEntry = namedtuple("FilingEntry", [
            "cik", "company_name", "form_type", "date_filed", "file_name",
            "path", "num_previously_valid"
        ])
        # idx file will have lines of the form CIK|Company Name|Form Type|Date Filed|File Name
        current_count = 0
        entries = re.findall(r'^[0-9]+[|].+[|].+[|][0-9\-]+[|].+$',
                             idx_file, re.MULTILINE)
        for entry in entries:
            fields = entry.split("|")
            fields[-1] = fields[-1].strip()  # get rid of carriage return or newline
            path = "Archives/{file_name}".format(file_name=fields[-1])
            entry = FilingEntry(*fields,
                                path=path,
                                num_previously_valid=current_count)
            if self.entry_filter is not None and not self.entry_filter(
                    entry):
                continue
            current_count += 1
            # Add new filing entry to CIK's list
            if entry.cik in self._filings_dict:
                self._filings_dict[entry.cik].append(entry)
            else:
                self._filings_dict[entry.cik] = [entry]
        return self._filings_dict

    def get_urls(self):
        """Get all URLs for day.

        Expects client _BASE to have trailing "/" for final URLs.

        Returns:
            urls (list of str): List of all URLs to get.
        """
        filings_dict = self.get_filings_dict()
        self._urls = {
            company:
            [self.client._prepare_query(entry.path) for entry in entries]
            for company, entries in filings_dict.items()
        }
        return self._urls

    @staticmethod
    def _do_create_and_copy(q):
        """Create path and copy file to end of path.

        Args:
            q (Queue.queue): Queue to get filename, new directory,
                and old path information from.
        """
        while True:
            try:
                filename, new_dir, old_path = q.get(timeout=1)
            except Empty:
                return
            make_path(new_dir)
            path = os.path.join(new_dir, filename)
            shutil.copyfile(old_path, path)
            q.task_done()

    @staticmethod
    def _do_unpack_archive(q, extract_directory):
        """Unpack archive file in given extract directory.

        Args:
            q (Queue.queue): Queue to get filname from.
            extract_directory (): Where to extract archive file.
        """
        while True:
            try:
                filename = q.get(timeout=1)
            except Empty:
                return
            shutil.unpack_archive(filename, extract_directory)
            os.remove(filename)
            q.task_done()

    def _unzip(self, extract_directory):
        """Unzips files from tar files into extract directory.

        Args:
            extract_directory (str): Temporary path to extract files to.
                Note that this directory will be completely removed after
                files are unzipped.
        """
        # Download tar files asynchronously into extract_directory
        tar_urls = self._get_tar_urls()
        inputs = [(url, os.path.join(extract_directory, url.split('/')[-1])) for url in tar_urls]
        asyncio.run(self.client.wait_for_download_async(inputs))

        tar_files = [p for _, p in inputs]

        unpack_queue = Queue(maxsize=len(tar_files))
        unpack_threads = len(tar_files)

        for _ in range(unpack_threads):
            worker = Thread(target=self._do_unpack_archive,
                            args=(unpack_queue, extract_directory))
            worker.start()
        for f in tar_files:
            full_path = os.path.join(extract_directory, f)
            unpack_queue.put_nowait(full_path)

        unpack_queue.join()

    def _move_to_dest(self, urls, extract_directory, directory, file_pattern,
                      dir_pattern):
        """Moves all files from extract_directory into proper final format in directory.

        Args:
            urls (dict): Dictionary of URLs that were retrieved. See
            ``secedgar.core._index.IndexFilings.get_urls()`` for more.
            extract_directory (str): Location of extract directory as used in
                ``secedgar.core.daily.DailyFilings._unzip``
            directory (str): Final parent directory to move files to.
            dir_pattern (str): Format string for subdirectories. Default is `{cik}`.
                Valid options are `cik`. See ``save`` method for more.
            file_pattern (str): Format string for files. Default is `{accession_number}`.
                Valid options are `accession_number`. See ``save`` method for more.
        """
        # Allocate threads to move files according to pattern
        link_list = [item for links in urls.values() for item in links]

        move_queue = Queue(maxsize=len(link_list))
        move_threads = 64
        for _ in range(move_threads):
            worker = Thread(target=self._do_create_and_copy, args=(move_queue,))
            worker.start()

        (_, _, extracted_files) = next(os.walk(extract_directory))

        for link in link_list:
            link_cik = link.split('/')[-2]
            link_accession = self.get_accession_number(link)
            filepath = link_accession.split('.')[0]
            possible_endings = ('nc', 'corr04', 'corr03', 'corr02', 'corr01')
            for ending in possible_endings:
                full_filepath = filepath + '.' + ending
                # If the filepath is found, move it to the correct path
                if full_filepath in extracted_files:

                    formatted_dir = dir_pattern.format(cik=link_cik)
                    formatted_file = file_pattern.format(
                        accession_number=link_accession)
                    old_path = os.path.join(extract_directory, full_filepath)
                    full_dir = os.path.join(directory, formatted_dir)
                    move_queue.put_nowait((formatted_file, full_dir, old_path))
                    break
        move_queue.join()

    def _save_filings(self,
                      directory,
                      dir_pattern="{cik}",
                      file_pattern="{accession_number}",
                      download_all=False,
                      **kwargs):
        """Save all filings.

        Will store all filings under the parent directory of ``directory``, further
        separating filings using ``dir_pattern`` and ``file_pattern``.

        Args:
            directory (str): Directory where filings should be stored.
            dir_pattern (str): Format string for subdirectories. Default is `{cik}`.
                Valid options are `{cik}`.
            file_pattern (str): Format string for files. Default is `{accession_number}`.
                Valid options are `{accession_number}`.
            download_all (bool): Type of downloading system, if true downloads all tar files,
                if false downloads each file in index. Default is `False`.
        """
        urls = self.get_urls_safely(**kwargs)

        if download_all:
            with tempfile.TemporaryDirectory() as tmpdir:
                self._unzip(extract_directory=tmpdir)
                # Apply folder structure by moving to final directory
                self._move_to_dest(urls=urls,
                                   extract_directory=tmpdir,
                                   directory=directory,
                                   file_pattern=file_pattern,
                                   dir_pattern=dir_pattern)
        else:
            inputs = []
            for company, links in urls.items():
                formatted_dir = dir_pattern.format(cik=company)
                for link in links:
                    formatted_file = file_pattern.format(
                        accession_number=self.get_accession_number(link))
                    path = os.path.join(directory, formatted_dir,
                                        formatted_file)
                    inputs.append((link, path))
            asyncio.run(self.client.wait_for_download_async(inputs))
